#!/usr/bin/env python
"""
Tool to automatically replace "import *" imports with explicit imports

Requires pyflakes.

Usage:

$ removestar file.py # Shows diff but does not edit file.py

$ removestar -i file.py # Edits file.py in-place

$ removestar -i module/ # Modifies every Python file in module/ recursively

"""
import argparse
import glob
import io
import os
import sys
import tempfile

import nbformat
from nbconvert import PythonExporter

from . import __version__
from .helper import get_diff_text
from .output import get_colored_diff, red
from .removestar import fix_code, replace_in_nb


class RawDescriptionHelpArgumentDefaultsHelpFormatter(
    argparse.RawDescriptionHelpFormatter, argparse.ArgumentDefaultsHelpFormatter
):
    pass


def main():  # noqa: PLR0912, C901
    parser = argparse.ArgumentParser(
        description=__doc__,
        prog="removestar",
        formatter_class=RawDescriptionHelpArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "paths", nargs="*", help="Files or directories to fix", metavar="PATH"
    )
    parser.add_argument(
        "-i", "--in-place", action="store_true", help="Edit the files in-place."
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + __version__,
        help="Show removestar version number and exit.",
    )
    parser.add_argument(
        "--no-skip-init",
        action="store_false",
        dest="skip_init",
        help="Don't skip __init__.py files (they are skipped by default)",
    )
    parser.add_argument(
        "--no-dynamic-importing",
        action="store_false",
        dest="allow_dynamic",
        help="""Don't dynamically import modules to determine the list of names. This is required for star imports from external modules and modules in the standard library.""",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="""Print information about every imported name that is replaced.""",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="""Don't print any warning messages.""",
    )
    parser.add_argument(
        "--max-line-length",
        type=int,
        default=100,
        help="""The maximum line length for replaced imports before they are wrapped. Set to 0 to disable line wrapping.""",
    )
    # For testing
    parser.add_argument("--_this-file", action="store_true", help=argparse.SUPPRESS)

    args = parser.parse_args()

    if args._this_file:
        print(__file__, end="")
        return

    if args.max_line_length == 0:
        args.max_line_length = float("inf")

    exit_1 = False
    for file in _iter_paths(args.paths):
        _, filename = os.path.split(file)
        if args.skip_init and filename == "__init__.py":
            continue

        if not os.path.isfile(file):
            print(red(f"Error: {file}: no such file or directory"), file=sys.stderr)
            continue
        if file.endswith(".py"):
            with open(file, encoding="utf-8") as f:
                code = f.read()

            try:
                new_code = fix_code(
                    code,
                    file=file,
                    max_line_length=args.max_line_length,
                    verbose=args.verbose,
                    quiet=args.quiet,
                    allow_dynamic=args.allow_dynamic,
                )
            except (RuntimeError, NotImplementedError) as e:
                if not args.quiet:
                    print(red(f"Error with {file}: {e}"), file=sys.stderr)
                continue

            if new_code != code:
                exit_1 = True
                if args.in_place:
                    with open(file, "w", encoding="utf-8") as f:
                        f.write(new_code)
                    if not args.quiet:
                        print(
                            get_colored_diff(
                                get_diff_text(
                                    io.StringIO(code).readlines(),
                                    io.StringIO(new_code).readlines(),
                                    file,
                                )
                            )
                        )
                else:
                    print(
                        get_colored_diff(
                            get_diff_text(
                                io.StringIO(code).readlines(),
                                io.StringIO(new_code).readlines(),
                                file,
                            )
                        )
                    )
        elif file.endswith(".ipynb"):
            tmp_file = tempfile.NamedTemporaryFile()
            tmp_path = tmp_file.name

            with open(file) as f:
                nb = nbformat.reads(f.read(), nbformat.NO_CONVERT)

            ## save as py
            exporter = PythonExporter()
            code, _ = exporter.from_notebook_node(nb)
            tmp_file.write(code.encode("utf-8"))

            try:
                new_code = fix_code(
                    code=code,
                    file=tmp_path,
                    max_line_length=args.max_line_length,
                    verbose=args.verbose,
                    quiet=args.quiet,
                    allow_dynamic=args.allow_dynamic,
                    return_replacements=True,
                )
                new_code_not_dict = fix_code(
                    code=code,
                    file=tmp_path,
                    max_line_length=args.max_line_length,
                    verbose=False,
                    quiet=True,
                    allow_dynamic=args.allow_dynamic,
                    return_replacements=False,
                )
            except (RuntimeError, NotImplementedError) as e:
                if not args.quiet:
                    print(red(f"Error with {file}: {e}"), file=sys.stderr)
                continue

            tmp_file.close()

            if new_code_not_dict != code:
                exit_1 = True
                if args.in_place:
                    with open(file) as f:
                        nb = nbformat.reads(f.read(), nbformat.NO_CONVERT)
                        fixed_code = replace_in_nb(
                            nb,
                            new_code,
                            cell_type="code",
                        )

                    with open(file, "w+") as f:
                        f.writelines(fixed_code)

                    if not args.quiet:
                        print(
                            get_colored_diff(
                                get_diff_text(
                                    io.StringIO(code).readlines(),
                                    io.StringIO(new_code_not_dict).readlines(),
                                    file,
                                )
                            )
                        )
                else:
                    print(
                        get_colored_diff(
                            get_diff_text(
                                io.StringIO(code).readlines(),
                                io.StringIO(new_code_not_dict).readlines(),
                                file,
                            )
                        )
                    )

    if exit_1:
        sys.exit(1)


def _iter_paths(paths):
    for path in paths:
        if os.path.isdir(path):
            for file in glob.iglob(path + "/**", recursive=True):
                if not file.endswith(".py") and not file.endswith(".ipynb"):
                    continue
                yield file
        else:
            yield path


if __name__ == "__main__":
    main()



def bold(line):
    return "\033[1m" + line + "\033[0m"  # bold, reset


def red(line):
    return "\033[31m" + line + "\033[0m"  # red, reset


def yellow(line):
    return "\033[33m" + line + "\033[0m"  # yellow, reset


def cyan(line):
    return "\033[36m" + line + "\033[0m"  # cyan, reset


def green(line):
    return "\033[32m" + line + "\033[0m"  # green, reset


def get_colored_diff(contents):
    """Inject the ANSI color codes to the diff."""
    # taken from https://github.com/psf/black/blob/main/src/black/output.py
    # Copyright (c) 2018 Łukasz Langa

    # Permission is hereby granted, free of charge, to any person obtaining a copy
    # of this software and associated documentation files (the "Software"), to deal
    # in the Software without restriction, including without limitation the rights
    # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    # copies of the Software, and to permit persons to whom the Software is
    # furnished to do so, subject to the following conditions:

    # The above copyright notice and this permission notice shall be included in all
    # copies or substantial portions of the Software.

    # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    # SOFTWARE.
    lines = contents.split("\n")
    for i, line in enumerate(lines):
        if line.startswith(("+++", "---")):
            line = bold(line)  # bold, reset  # noqa: PLW2901
        elif line.startswith("@@"):
            line = cyan(line)  # cyan, reset  # noqa: PLW2901
        elif line.startswith("+"):
            line = green(line)  # green, reset  # noqa: PLW2901
        elif line.startswith("-"):
            line = red(line)  # red, reset  # noqa: PLW2901
        lines[i] = line
    return "\n".join(lines)
